// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0

import Foundation
import UIKit
import SnapKit
import Shared

/* The layout for ETP Cover Sheet
    
 |----------------|
 |            Done|
 |                |
 |     Image      |
 |    [Centre]    | (Top View)
 |                |
 |Title Multiline |
 |                | 
 |Description     |
 |Multiline       |
 |                |
 |----------------|
 |                |
 |    [Button]    |
 |    [Button]    | (Bottom View)
 |----------------|
 
 */

class ETPCoverSheetViewController: UIViewController {

    struct UX {
        static let doneButtonPadding: CGFloat = 20
        static let doneButtonHeight: CGFloat = 20
        static let primaryButtonColour = UIColor.Photon.Blue50
        static let primaryButtonCornerRadius: CGFloat = 10
        static let primaryButtonFont = UIFont.systemFont(ofSize: 18, weight: .semibold)
        static let primaryButtonHeight: CGFloat = 46
        static let primaryButtonEdgeInset: CGFloat = 18
    }

    // Public constants
    let viewModel = ETPViewModel()
    static let theme = BuiltinThemeName(rawValue: LegacyThemeManager.instance.current.name) ?? .normal
    // Private vars
    private var fxTextThemeColour: UIColor {
        // For dark theme we want to show light colours and for light we want to show dark colours
        return ETPCoverSheetViewController.theme == .dark ? .white : .black
    }
    private var fxBackgroundThemeColour: UIColor {
        return ETPCoverSheetViewController.theme == .dark ? .black : .white
    }
    private var doneButton: UIButton = {
        let button = UIButton()
        button.setTitle(.SettingsSearchDoneButton, for: .normal)
        button.titleLabel?.font = UIFont.systemFont(ofSize: 18, weight: .regular)
        button.setTitleColor(UIColor.systemBlue, for: .normal)
        return button
    }()
    private lazy var topImageView: UIImageView = {
        let imgView = UIImageView(image: viewModel.etpCoverSheetmodel?.titleImage)
        imgView.contentMode = .scaleAspectFit
        imgView.clipsToBounds = true
        return imgView
    }()
    private lazy var titleLabel: UILabel = {
        let label = UILabel()
        label.text = viewModel.etpCoverSheetmodel?.titleText
        label.textColor = fxTextThemeColour
        label.font = UIFont.boldSystemFont(ofSize: 18)
        label.textAlignment = .left
        label.numberOfLines = 0
        return label
    }()
    private lazy var descriptionLabel: UILabel = {
        let label = UILabel()
        label.text = viewModel.etpCoverSheetmodel?.descriptionText
        label.textColor = fxTextThemeColour
        label.font = UIFont.systemFont(ofSize: 18)
        label.textAlignment = .left
        label.numberOfLines = 0
        return label
    }()
    private lazy var goToSettingsButton: UIButton = {
        let button = UIButton()
        button.setTitle(.CoverSheetETPSettingsButton, for: .normal)
        button.titleLabel?.font = UX.primaryButtonFont
        button.layer.cornerRadius = UX.primaryButtonCornerRadius
        button.setTitleColor(.white, for: .normal)
        button.backgroundColor = UX.primaryButtonColour
        return button
    }()
    private lazy var startBrowsingButton: UIButton = {
        let button = UIButton()
        button.setTitle(.StartBrowsingButtonTitle, for: .normal)
        button.titleLabel?.font = UX.primaryButtonFont
        button.setTitleColor(UIColor.Photon.Blue50, for: .normal)
        button.backgroundColor = .clear
        return button
    }()

    init() {
        super.init(nibName: nil, bundle: nil)
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func viewDidLoad() {
        super.viewDidLoad()
        initialViewSetup()
    }

    func initialViewSetup() {
        self.view.backgroundColor = fxBackgroundThemeColour

        // Initialize
        self.view.addSubview(topImageView)
        self.view.addSubview(doneButton)
        self.view.addSubview(titleLabel)
        self.view.addSubview(descriptionLabel)
        self.view.addSubview(goToSettingsButton)
        self.view.addSubview(startBrowsingButton)

        // Constraints
        setupTopView()
        setupBottomView()
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
    }

    private func setupTopView() {
        // Done button target setup
        doneButton.addTarget(self, action: #selector(dismissAnimated), for: .touchUpInside)
        // Done button constraints setup
        // This button is located at top right hence top, right and height
        doneButton.snp.makeConstraints { make in
            make.top.equalTo(view.snp.topMargin).offset(UX.doneButtonPadding)
            make.right.equalToSuperview().inset(-UX.doneButtonPadding)
            make.height.equalTo(UX.doneButtonHeight)
        }
        // The top imageview constraints setup
        topImageView.snp.makeConstraints { make in
            make.left.equalToSuperview()
            make.right.equalToSuperview()
            make.top.equalTo(doneButton.snp.bottom).offset(10)
            make.bottom.equalTo(goToSettingsButton.snp.top).offset(-200)
            make.height.lessThanOrEqualTo(100)
        }
        // Top title label constraints setup
        titleLabel.snp.makeConstraints { make in
            make.bottom.equalTo(descriptionLabel.snp.top).offset(-5)
            make.left.right.equalToSuperview().inset(20)
        }
        // Description title label constraints setup
        descriptionLabel.snp.makeConstraints { make in
            make.bottom.equalTo(goToSettingsButton.snp.top).offset(-40)
            make.left.right.equalToSuperview().inset(20)
        }
    }

    private func setupBottomView() {
        // Bottom start button constraints
        goToSettingsButton.snp.makeConstraints { make in
            make.left.right.equalToSuperview().inset(UX.primaryButtonEdgeInset)
            make.bottom.equalTo(startBrowsingButton.snp.top).offset(-16)
            make.height.equalTo(UX.primaryButtonHeight)
        }
        // Bottom goto settings button
        goToSettingsButton.addTarget(self, action: #selector(goToSettings), for: .touchUpInside)
        // Bottom start button constraints
        startBrowsingButton.snp.makeConstraints { make in
            make.left.right.equalToSuperview().inset(UX.primaryButtonEdgeInset)
            let h = view.frame.height
            // On large iPhone screens, bump this up from the bottom
            let offset: CGFloat = UIDevice.current.userInterfaceIdiom == .pad ? 20 : (h > 800 ? 60 : 20)
            make.bottom.equalTo(view.safeArea.bottom).inset(offset)
            make.height.equalTo(UX.primaryButtonHeight)
        }
        // Bottom start browsing target setup
        startBrowsingButton.addTarget(self, action: #selector(startBrowsing), for: .touchUpInside)
    }

    // Button Actions
    @objc private func dismissAnimated() {
        self.dismiss(animated: true, completion: nil)
        TelemetryWrapper.recordEvent(category: .action, method: .press, object: .dismissedETPCoverSheet)
    }

    @objc private func goToSettings() {
        viewModel.goToSettings?()
        TelemetryWrapper.recordEvent(category: .action, method: .press, object: .dismissETPCoverSheetAndGoToSettings)
    }

    @objc private func startBrowsing() {
        viewModel.startBrowsing?()
        TelemetryWrapper.recordEvent(category: .action, method: .press, object: .dismissUpdateCoverSheetAndStartBrowsing)
    }
}

// UIViewController setup to keep it in portrait mode
extension ETPCoverSheetViewController {
    override var shouldAutorotate: Bool {
        return false
    }

    override var supportedInterfaceOrientations: UIInterfaceOrientationMask {
        // This actually does the right thing on iPad where the modally
        // presented version happily rotates with the iPad orientation.
        return .portrait
    }
}
